@@ -3,13 +3,14 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-DEFAULT_SSH_CMD = "ssh"
-GOT_PASSWORD_PROMPT_SUGGESTION = "Perhaps you need to set up your ssh agent?\n"
 SYSTEM_TRAY_SUPPORTED = True
+REINIT_WINDOWS = True
+
 CLIPBOARDS=["CLIPBOARD"]
 CLIPBOARD_WANT_TARGETS = True
 CLIPBOARD_GREEDY = True
-CLIPBOARD_NATIVE_CLASS = ("xpra.platform.darwin.osx_clipboard", "OSXClipboardProtocolHelper", {})
-SHADOW_SUPPORTED = True
-CAN_DAEMONIZE = True
+CLIPBOARD_NATIVE_CLASS = "xpra.platform.darwin.osx_clipboard.OSXClipboardProtocolHelper"
+
 UI_THREAD_POLLING = 500    #poll every 500 ms
+
+DEFAULT_SSH_COMMAND = "ssh"
