@@ -19,10 +19,9 @@ import shlex
 import traceback
 
 from xpra import __version__ as XPRA_VERSION
-from xpra.dotxpra import DotXpra
+from xpra.platform.dotxpra import DotXpra
 from xpra.platform.features import LOCAL_SERVERS_SUPPORTED, SHADOW_SUPPORTED, CAN_DAEMONIZE
 from xpra.platform.options import add_client_options
-from xpra.net.crypto import ENCRYPTION_CIPHERS
 from xpra.util import csv
 from xpra.scripts.config import OPTION_TYPES, \
     InitException, InitInfo, InitExit, \
@@ -30,6 +29,7 @@ from xpra.scripts.config import OPTION_TYPES, \
     make_defaults_struct, parse_bool, print_bool, print_number, validate_config, has_sound_support, name_to_field
 
 
+VSOCK_TIMEOUT = int(os.environ.get("XPRA_VSOCK_TIMEOUT", 5))
 SOCKET_TIMEOUT = int(os.environ.get("XPRA_SOCKET_TIMEOUT", 10))
 TCP_NODELAY = int(os.environ.get("XPRA_TCP_NODELAY", "1"))
 NO_ROOT_WARNING = int(os.environ.get("XPRA_NO_ROOT_WARNING", "0"))
@@ -63,12 +63,14 @@ def warn(msg):
     sys.stderr.write(msg+"\n")
 
 def nox():
-    if "DISPLAY" in os.environ:
+    DISPLAY = os.environ.get("DISPLAY")
+    if DISPLAY is not None:
         del os.environ["DISPLAY"]
     # This is an error on Fedora/RH, so make it an error everywhere so it will
     # be noticed:
     import warnings
     warnings.filterwarnings("error", "could not open display")
+    return DISPLAY
 
 
 supports_shadow = SHADOW_SUPPORTED
@@ -242,7 +244,22 @@ def do_parse_cmdline(cmdline, defaults):
                       help="program to spawn in new server (may be repeated). Default: %default.")
     group.add_option("--start-child", action="append",
                       dest="start_child", metavar="CMD", default=list(defaults.start_child or []),
-                      help="program to spawn in new server, taken into account by the exit-with-children option (may be repeated). Default: %default.")
+                      help="program to spawn in new server, taken into account by the exit-with-children option (may be repeated to run multiple commands). Default: %default.")
+    group.add_option("--start-after-connect", action="append",
+                      dest="start_after_connect", default=defaults.start_after_connect,
+                      help="program to spawn in new server after the first client has connected (may be repeated). Default: %default.")
+    group.add_option("--start-child-after-connect", action="append",
+                      dest="start_child_after_connect", default=defaults.start_child_after_connect,
+                      help="program to spawn in new server after the first client has connected, taken into account by the exit-with-children option (may be repeated to run multiple commands). Default: %default.")
+    group.add_option("--start-on-connect", action="append",
+                      dest="start_on_connect", default=defaults.start_on_connect,
+                      help="program to spawn in new server every time a client connects (may be repeated). Default: %default.")
+    group.add_option("--start-child-on-connect", action="append",
+                      dest="start_child_on_connect", default=defaults.start_child_on_connect,
+                      help="program to spawn in new server every time a client connects, taken into account by the exit-with-children option (may be repeated). Default: %default.")
+    group.add_option("--exec-wrapper", action="store",
+                      dest="exec_wrapper", metavar="CMD", default=defaults.exec_wrapper,
+                      help="Wrapper for executing commands. Default: %default.")
     group.add_option("--exit-with-children", action="store_true",
                       dest="exit_with_children", default=defaults.exit_with_children,
                       help="Terminate the server when the last --start-child command(s) exit")
@@ -250,6 +267,9 @@ def do_parse_cmdline(cmdline, defaults):
     group.add_option("--start-new-commands", action="store", metavar="yes|no",
                       dest="start_new_commands", default=defaults.start_new_commands,
                       help="Allows clients to execute new commands on the server. Default: %s." % enabled_str(defaults.start_new_commands))
+    group.add_option("--dbus-launch", action="store",
+                      dest="dbus_launch", metavar="CMD", default=defaults.dbus_launch,
+                      help="Start the session within a dbus-launch context, leave empty to turn off. Default: %default.")
     group.add_option("--env", action="append",
                       dest="env", default=list(defaults.env or []),
                       help="Define environment variables used with 'start-child' and 'start', can be specified multiple times. Default: %s." % ", ".join([("'%s'" % x) for x in (defaults.env or []) if not x.startswith("#")]))
@@ -346,14 +366,33 @@ def do_parse_cmdline(cmdline, defaults):
         ignore({"use-display"   : False,
                 "xvfb"          : '',
                 "fake-xinerama" : defaults.fake_xinerama})
+    group.add_option("--resize-display", action="store",
+                      dest="resize_display", default=defaults.resize_display, metavar="yes|no",
+                      help="Whether the server display should be resized to match the client resolution. Default: %s." % enabled_str(defaults.resize_display))
     if supports_server or supports_shadow:
+        group.add_option("--bind", action="append",
+                          dest="bind", default=list(defaults.bind or []),
+                          metavar="SOCKET",
+                          help="Listen for connections over unix domain sockets. You may specify this option multiple times to listen on different locations.")
         group.add_option("--bind-tcp", action="append",
                           dest="bind_tcp", default=list(defaults.bind_tcp or []),
                           metavar="[HOST]:PORT",
-                          help="Listen for connections over TCP (use --password-file to secure it)."
+                          help="Listen for connections over TCP (use --tcp-auth to secure it)."
                             + " You may specify this option multiple times with different host and port combinations")
     else:
         ignore({"bind-tcp" : []})
+    try:
+        from xpra.net import vsock
+    except:
+        vsock = None
+    if vsock:
+        group.add_option("--bind-vsock", action="append",
+                          dest="bind_vsock", default=list(defaults.bind_vsock or []),
+                          metavar="[CID]:[PORT]",
+                          help="Listen for connections over VSOCK."
+                            + " You may specify this option multiple times with different CID and port combinations")
+    else:
+        ignore({"bind-vsock" : []})
     legacy_bool_parse("mdns")
     if (supports_server or supports_shadow):
         group.add_option("--mdns", action="store", metavar="yes|no",
@@ -371,6 +410,9 @@ def do_parse_cmdline(cmdline, defaults):
         group.add_option("--pulseaudio-command", action="store",
                       dest="pulseaudio_command", default=defaults.pulseaudio_command,
                       help="The command used to start the pulseaudio server. Default: '%default'.")
+        group.add_option("--pulseaudio-configure-commands", action="append",
+                      dest="pulseaudio_configure_commands", default=defaults.pulseaudio_configure_commands,
+                      help="The commands used to configure the pulseaudio server. Default: '%default'.")
         group.add_option("--dbus-proxy", action="store", metavar="yes|no",
                       dest="dbus_proxy", default=defaults.dbus_proxy,
                       help="Forward dbus calls from the client. Default: %s." % enabled_str(defaults.dbus_proxy))
@@ -382,6 +424,7 @@ def do_parse_cmdline(cmdline, defaults):
                 "pulseaudio-command"    : defaults.pulseaudio_command,
                 "dbus-proxy"            : defaults.dbus_proxy,
                 "dbus-control"          : defaults.dbus_control,
+                "pulseaudio-configure-commands" : defaults.pulseaudio_configure_commands,
                 })
 
     group = optparse.OptionGroup(parser, "Server Controlled Features",
@@ -409,6 +452,10 @@ def do_parse_cmdline(cmdline, defaults):
     group.add_option("--bell", action="store",
                       dest="bell", default=defaults.bell, metavar="yes|no",
                       help="Forward the system bell. Default: %s." % enabled_str(defaults.bell))
+    legacy_bool_parse("webcam")
+    group.add_option("--webcam", action="store",
+                      dest="webcam", default=defaults.webcam,
+                      help="Webcam forwarding, can be used to specify a device. Default: %s." % defaults.webcam)
     legacy_bool_parse("global-menus")
     group.add_option("--global-menus", action="store",
                       dest="global_menus", default=defaults.global_menus, metavar="yes|no",
@@ -432,7 +479,7 @@ def do_parse_cmdline(cmdline, defaults):
     group.add_option("--sharing", action="store", metavar="yes|no",
                       dest="sharing", default=defaults.sharing,
                       help="Allow more than one client to connect to the same session. Default: %s." % enabled_str(defaults.sharing))
-    group.add_option("--remote-logging", action="store", metavar="yes|no",
+    group.add_option("--remote-logging", action="store", metavar="yes|no|both",
                       dest="remote_logging", default=defaults.remote_logging,
                       help="Forward all the client's log output to the server. Default: %s." % enabled_str(defaults.remote_logging))
     legacy_bool_parse("speaker")
@@ -464,9 +511,9 @@ def do_parse_cmdline(cmdline, defaults):
                           help="Try to synchronize sound and video. Default: %s." % enabled_str(defaults.av_sync))
     else:
         ignore({"av-sync"           : False,
-                "speaker"           : False,
+                "speaker"           : "no",
                 "speaker-codec"     : [],
-                "microphone"        : False,
+                "microphone"        : "no",
                 "microphone-codec"  : [],
                 "sound-source"      : ""})
 
@@ -578,6 +625,9 @@ def do_parse_cmdline(cmdline, defaults):
     group.add_option("--window-icon", action="store",
                           dest="window_icon", default=defaults.window_icon,
                           help="Path to the default image which will be used for all windows (the application may override this)")
+    group.add_option("--window-close", action="store",
+                          dest="window_close", default=defaults.window_close,
+                          help="The action to take when a window is closed by the client. Valid options are: 'forward', 'ignore', 'disconnect'. Default: '%default'.")
     # let the platform specific code add its own options:
     # adds "--no-tray" for platforms that support it
     add_client_options(cmdline, group, defaults)
@@ -637,6 +687,12 @@ def do_parse_cmdline(cmdline, defaults):
     group.add_option("--tcp-auth", action="store",
                       dest="tcp_auth", default=defaults.tcp_auth,
                       help="The authentication module to use for TCP sockets (default: '%default')")
+    if vsock:
+        group.add_option("--vsock-auth", action="store",
+                         dest="vsock_auth", default=defaults.vsock_auth,
+                         help="The authentication module to use for vsock sockets (default: '%default')")
+    else:
+        ignore({"vsock-auth" : ""})
     if os.name=="posix":
         group.add_option("--mmap-group", action="store_true",
                           dest="mmap_group", default=defaults.mmap_group,
@@ -666,11 +722,11 @@ def do_parse_cmdline(cmdline, defaults):
     group.add_option("--remote-xpra", action="store",
                       dest="remote_xpra", default=defaults.remote_xpra,
                       metavar="CMD",
-                      help="How to run xpra on the remote host (default: '%default')")
+                      help="How to run xpra on the remote host (default: %s)" % (" or ".join(defaults.remote_xpra)))
     group.add_option("--encryption", action="store",
                       dest="encryption", default=defaults.encryption,
                       metavar="ALGO",
-                      help="Specifies the encryption cipher to use, supported algorithms are: %s (default: None)" % (", ".join(ENCRYPTION_CIPHERS) or 'none available'))
+                      help="Specifies the encryption cipher to use, specify 'help' to get a list of options. (default: None)")
     group.add_option("--encryption-keyfile", action="store",
                       dest="encryption_keyfile", default=defaults.encryption_keyfile,
                       metavar="FILE",
@@ -678,7 +734,7 @@ def do_parse_cmdline(cmdline, defaults):
     group.add_option("--tcp-encryption", action="store",
                       dest="tcp_encryption", default=defaults.tcp_encryption,
                       metavar="ALGO",
-                      help="Specifies the encryption cipher to use for TCP sockets, supported algorithms are: %s (default: None)" % (", ".join(ENCRYPTION_CIPHERS) or 'none available'))
+                      help="Specifies the encryption cipher to use for TCP sockets, specify 'help' to get a list of options. (default: None)")
     group.add_option("--tcp-encryption-keyfile", action="store",
                       dest="tcp_encryption_keyfile", default=defaults.tcp_encryption_keyfile,
                       metavar="FILE",
@@ -686,6 +742,11 @@ def do_parse_cmdline(cmdline, defaults):
 
     options, args = parser.parse_args(cmdline[1:])
 
+    #remote-xpra is meant to be a list, but the user can specify a string using the command line,
+    #in which case we replace all the default values with this single entry:
+    if not isinstance(options.remote_xpra, (list, tuple)):
+        options.remote_xpra = [options.remote_xpra]
+
     #ensure all the option fields are set even though
     #some options are not shown to the user:
     for k,v in hidden_options.items():
@@ -773,24 +834,49 @@ def do_parse_cmdline(cmdline, defaults):
         except:
             raise InitException("invalid max-size: %s" % options.max_size)
         options.max_size = "%sx%s" % (w, h)
-    if options.encryption or options.tcp_encryption:
-        if not ENCRYPTION_CIPHERS:
-            raise InitException("cannot use encryption: no ciphers available (pycrypto must be installed)")
-        if options.encryption and options.encryption not in ENCRYPTION_CIPHERS:
-            raise InitException("encryption %s is not supported, try: %s" % (options.encryption, ", ".join(ENCRYPTION_CIPHERS)))
-        if options.tcp_encryption and options.tcp_encryption not in ENCRYPTION_CIPHERS:
-            raise InitException("encryption %s is not supported, try: %s" % (options.tcp_encryption, ", ".join(ENCRYPTION_CIPHERS)))
-        #password file can be used as fallback for encryption keys:
-        has_key = options.password_file or os.environ.get('XPRA_PASSWORD') or os.environ.get('XPRA_ENCRYPTION_KEY')
-        if not has_key:
-            if options.encryption and not options.encryption_keyfile:
-                raise InitException("encryption %s cannot be used without a keyfile (see --encryption-keyfile option)" % options.encryption)
-            if options.tcp_encryption and not options.tcp_encryption_keyfile:
-                raise InitException("tcp-encryption %s cannot be used without a keyfile (see --tcp-encryption-keyfile option)" % options.tcp_encryption)
+    if options.encryption_keyfile and not options.encryption:
+        options.encryption = "AES"
+    if options.tcp_encryption_keyfile and not options.tcp_encryption:
+        options.tcp_encryption = "AES"
     #ensure opengl is either True, False or None
     options.opengl = parse_bool("opengl", options.opengl)
     return options, args
 
+def validate_encryption(opts):
+    do_validate_encryption(opts.auth, opts.tcp_auth, opts.encryption, opts.tcp_encryption, opts.password_file, opts.encryption_keyfile, opts.tcp_encryption_keyfile)
+
+def do_validate_encryption(auth, tcp_auth, encryption, tcp_encryption, password_file, encryption_keyfile, tcp_encryption_keyfile):
+    if not encryption and not tcp_encryption:
+        #don't bother initializing anything
+        return
+    from xpra.net.crypto import crypto_backend_init
+    crypto_backend_init()
+    if not encryption and not tcp_encryption:
+        return
+    env_key = os.environ.get("XPRA_ENCRYPTION_KEY")
+    pass_key = os.environ.get("XPRA_PASSWORD")
+    from xpra.net.crypto import ENCRYPTION_CIPHERS
+    if not ENCRYPTION_CIPHERS:
+        raise InitException("cannot use encryption: no ciphers available (a crypto library must be installed)")
+    if encryption=="help" or tcp_encryption=="help":
+        raise InitInfo("the following encryption ciphers are available: %s" % csv(ENCRYPTION_CIPHERS))
+    if encryption and encryption not in ENCRYPTION_CIPHERS:
+        raise InitException("encryption %s is not supported, try: %s" % (encryption, csv(ENCRYPTION_CIPHERS)))
+    if tcp_encryption and tcp_encryption not in ENCRYPTION_CIPHERS:
+        raise InitException("encryption %s is not supported, try: %s" % (tcp_encryption, csv(ENCRYPTION_CIPHERS)))
+    if encryption and not encryption_keyfile and not env_key and not auth:
+        raise InitException("encryption %s cannot be used without an authentication module or keyfile (see --encryption-keyfile option)" % encryption)
+    if tcp_encryption and not tcp_encryption_keyfile and not env_key and not tcp_auth:
+        raise InitException("tcp-encryption %s cannot be used without a tcp authentication module or keyfile (see --tcp-encryption-keyfile option)" % tcp_encryption)
+    if pass_key and env_key and pass_key==env_key:
+        raise InitException("encryption and authentication should not use the same value")
+    #discouraged but not illegal:
+    #if password_file and encryption_keyfile and password_file==encryption_keyfile:
+    #    if encryption:
+    #        raise InitException("encryption %s should not use the same file as the password authentication file" % encryption)
+    #    elif tcp_encryption:
+    #        raise InitException("tcp-encryption %s should not use the same file as the password authentication file" % tcp_encryption)
+
 def dump_frames(*arsg):
     frames = sys._current_frames()
     print("")
@@ -833,7 +919,7 @@ def show_sound_codec_help(is_server, speaker_codecs, microphone_codecs):
 
 def configure_logging(options, mode):
     to = sys.stderr
-    if mode in ("showconfig", "info", "control", "list", "attach", "stop", "version", "print"):
+    if mode in ("showconfig", "info", "control", "list", "attach", "stop", "version", "print", "opengl"):
         to = sys.stdout
     if mode in ("start", "upgrade", "attach", "shadow", "proxy", "_sound_record", "_sound_play", "stop", "print", "showconfig"):
         if "help" in options.speaker_codec or "help" in options.microphone_codec:
@@ -917,9 +1003,9 @@ def run_mode(script_file, error_cb, options, args, mode, defaults):
             #ie: "xpra start ssh:HOST:DISPLAY --start-child=xterm"
             return run_remote_server(error_cb, options, args, mode, defaults)
         elif (mode in ("start", "upgrade", "proxy") and supports_server) or (mode=="shadow" and supports_shadow):
-            nox()
+            current_display = nox()
             from xpra.scripts.server import run_server
-            return run_server(error_cb, options, mode, script_file, args)
+            return run_server(error_cb, options, mode, script_file, args, current_display)
         elif mode in ("attach", "detach", "screenshot", "version", "info", "control", "_monitor", "print"):
             return run_client(error_cb, options, args, mode)
         elif mode in ("stop", "exit") and (supports_server or supports_shadow):
@@ -935,10 +1021,12 @@ def run_mode(script_file, error_cb, options, args, mode, defaults):
                 error_cb("no sound support!")
             from xpra.sound.wrapper import run_sound
             return run_sound(mode, error_cb, options, args)
+        elif mode=="opengl":
+            return run_glcheck(options)
         elif mode == "initenv":
-            from xpra.scripts.server import xpra_runner_shell_script, write_runner_shell_script
+            from xpra.scripts.server import xpra_runner_shell_script, write_runner_shell_scripts
             script = xpra_runner_shell_script(script_file, os.getcwd(), options.socket_dir)
-            write_runner_shell_script(script, False)
+            write_runner_shell_scripts(script, False)
             return 0
         elif mode == "showconfig":
             return run_showconfig(options, args)
@@ -950,6 +1038,29 @@ def run_mode(script_file, error_cb, options, args, mode, defaults):
         return 128+signal.SIGINT
 
 
+def parse_vsock(vsock_str):
+    from xpra.net.vsock import STR_TO_CID, CID_ANY, PORT_ANY    #@UnresolvedImport
+    if not vsock_str.find(":")>=0:
+        raise InitException("invalid vsocket format '%s'" % vsock_str)
+    cid_str, port_str = vsock_str.split(":", 1)
+    if cid_str.lower() in ("auto", "any"):
+        cid = CID_ANY
+    else:
+        try:
+            cid = int(cid_str)
+        except ValueError:
+            cid = STR_TO_CID.get(cid_str.upper())
+            if cid is None:
+                raise InitException("invalid vsock cid '%s'" % cid_str)
+    if port_str.lower() in ("auto", "any"):
+        iport = PORT_ANY
+    else:
+        try:
+            iport = int(port_str)
+        except ValueError:
+            raise InitException("invalid vsock port '%s'" % port_str)
+    return cid, iport
+
 def parse_display_name(error_cb, opts, display_name):
     desc = {"display_name" : display_name}
     if display_name.lower().startswith("ssh:") or display_name.lower().startswith("ssh/"):
@@ -982,6 +1093,7 @@ def parse_display_name(error_cb, opts, display_name):
         #sys.platform.startswith("win")
         ssh_cmd = ssh[0].lower()
         is_putty = ssh_cmd.endswith("plink") or ssh_cmd.endswith("plink.exe")
+        desc["is_putty"] = is_putty
 
         upos = host.find("@")
         if upos>=0:
@@ -1040,17 +1152,16 @@ def parse_display_name(error_cb, opts, display_name):
                      "host"     : host,
                      "full_ssh" : full_ssh
                      })
-        remote_xpra = opts.remote_xpra.split()
+        desc["remote_xpra"] = opts.remote_xpra
         if opts.socket_dir:
-            #ie: XPRA_SOCKET_DIR=/tmp .xpra/run-xpra _proxy :10
-            remote_xpra.append("--socket-dir=%s" % opts.socket_dir)
-        desc["remote_xpra"] = remote_xpra
+            desc["socket_dir"] = opts.socket_dir
         if desc.get("password") is None and opts.password_file and os.path.exists(opts.password_file):
             try:
                 with open(opts.password_file, "rb") as f:
                     desc["password"] = f.read()
             except Exception as e:
-                print("failed to read password file %s: %s", opts.password_file, e)
+                sys.stderr.write("Error: failed to read the password file '%s':\n", opts.password_file)
+                sys.stderr.write(" %s\n", e)
         return desc
     elif display_name.startswith("socket:"):
         #use the socketfile specified:
@@ -1114,6 +1225,30 @@ def parse_display_name(error_cb, opts, display_name):
             host = "127.0.0.1"
         desc["host"] = host
         return desc
+    elif display_name.startswith("vsock:"):
+        #use the vsock specified:
+        vsock_str = display_name[len("vsock:"):]
+        cid, iport = parse_vsock(vsock_str)
+        desc.update({
+                "type"          : "vsock",
+                "local"         : False,
+                "display"       : display_name,
+                "vsock"         : (cid, iport),
+                })
+        opts.display = display_name
+        return desc
+    elif sys.platform.startswith("win") or display_name.startswith("named-pipe:"):
+        pipe_name = display_name
+        if display_name.startswith("named-pipe:"):
+            pipe_name = display_name[len("named-pipe:"):]
+        desc.update({
+                     "type"             : "named-pipe",
+                     "local"            : True,
+                     "display"          : "DISPLAY",
+                     "named-pipe"       : pipe_name,
+                     })
+        opts.display = display_name
+        return desc
     else:
         error_cb("unknown format for display name: %s" % display_name)
 
@@ -1122,15 +1257,23 @@ def pick_display(error_cb, opts, extra_args):
         # Pick a default server
         dotxpra = DotXpra(opts.socket_dir, opts.socket_dirs)
         dir_servers = dotxpra.socket_details(matching_state=DotXpra.LIVE)
-        sockdir, display, sockpath = single_display_match(dir_servers, error_cb)
-        desc = {
+        sockdir, display, path = single_display_match(dir_servers, error_cb)
+        if sys.platform.startswith("win"):
+            return {
+                    "type"              : "named-pipe",
+                    "local"             : True,
+                    "display"           : display,
+                    "display_name"      : display,
+                    "named-pipe"        : path,
+                    }
+        return {
                 "type"          : "unix-domain",
                 "local"         : True,
                 "display"       : display,
                 "display_name"  : display,
                 "socket_dir"    : sockdir,
-                "socket_path"   : sockpath}
-        return desc
+                "socket_path"   : path
+                }
     elif len(extra_args) == 1:
         return parse_display_name(error_cb, opts, extra_args[0])
     else:
@@ -1138,20 +1281,20 @@ def pick_display(error_cb, opts, extra_args):
 
 def single_display_match(dir_servers, error_cb):
     #ie: {"/tmp" : [LIVE, "desktop-10", "/tmp/desktop-10"]}
-    if len(dir_servers) == 0:
+    #aggregate all the different locations:
+    allservers = []
+    for sockdir, servers in dir_servers.items():
+        for state, name, path in servers:
+            if state==DotXpra.LIVE:
+                allservers.append((sockdir, name, path))
+    if len(allservers) == 0:
         error_cb("cannot find any live servers to connect to")
-    if len(dir_servers) > 1:
-        error_cb("there are multiple servers running in multiple locations, please specify")
-    assert len(dir_servers)==1, "len(dir_servers)=%s" % len(dir_servers)
-    #ie: [(LIVE, "desktop-10", "/tmp/desktop-10"), ]
-    sockdir, servers = list(dir_servers.items())[0]
-    assert len(servers)>0
-    if len(servers) > 1:
-        error_cb("there are multiple servers running in %s, please specify" % sockdir)
-    #ie: (LIVE, "desktop-10", "/tmp/desktop-10")
-    server = servers[0]
+    if len(allservers) > 1:
+        error_cb("there are multiple servers running, please specify")
+    assert len(allservers)==1, "len(dir_servers)=%s" % len(dir_servers)
+    sockdir, name, path = allservers[0]
     #ie: ("/tmp", "desktop-10", "/tmp/desktop-10")
-    return sockdir, server[1], server[2]
+    return sockdir, name, path
 
 
 def _socket_connect(sock, endpoint, description, dtype):
@@ -1169,8 +1312,6 @@ def connect_or_fail(display_desc):
     except InitException:
         raise
     except Exception as e:
-        raise
-    #    traceback.print_stack()
         raise InitException("connection failed: %s" % e)
 
 def ssh_connect_failed(message):
@@ -1185,16 +1326,10 @@ def connect_to(display_desc, debug_cb=None, ssh_fail_cb=ssh_connect_failed):
     conn = None
     if dtype == "ssh":
         from xpra.net import ConnectionClosedException
-        cmd = display_desc["full_ssh"]
-        proxy_cmd = display_desc["remote_xpra"] + display_desc["proxy_command"] + display_desc["display_as_args"]
-        if INITENV_COMMAND:
-            cmd += [INITENV_COMMAND+";"+(" ".join(proxy_cmd))]
-        else:
-            cmd = proxy_cmd
         try:
+            cmd = display_desc["full_ssh"]
             kwargs = {}
-            if "env" in display_desc:
-                kwargs["env"] = display_desc["env"]
+            env = display_desc.get("env")
             kwargs["stderr"] = sys.stderr
             if not display_desc.get("exit_ssh", False) and os.name=="posix" and not sys.platform.startswith("darwin"):
                 def setsid():
@@ -1206,6 +1341,24 @@ def connect_to(display_desc, debug_cb=None, ssh_fail_cb=ssh_connect_failed):
                 flags = CREATE_NEW_PROCESS_GROUP | CREATE_NEW_CONSOLE
                 kwargs["creationflags"] = flags
                 kwargs["stderr"] = PIPE
+            remote_xpra = display_desc["remote_xpra"]
+            assert len(remote_xpra)>0
+            remote_commands = []
+            socket_dir = display_desc.get("socket_dir")
+            for x in remote_xpra:
+                #ie: ["~/.xpra/run-xpra"] + ["_proxy"] + [":10"]
+                pc = [x] + display_desc["proxy_command"] + display_desc["display_as_args"]
+                if socket_dir:
+                    pc.append("--socket-dir=%s" % socket_dir)
+                remote_commands.append((" ".join(pc)))
+            #ie: ~/.xpra/run-xpra _proxy || $XDG_RUNTIME_DIR/run-xpra _proxy
+            remote_cmd = " || ".join(remote_commands)
+            if INITENV_COMMAND:
+                cmd += [INITENV_COMMAND+";"+remote_cmd]
+            else:
+                cmd += [remote_cmd]
+            if env:
+                kwargs["env"] = env
             if debug_cb:
                 debug_cb("starting %s tunnel" % str(cmd[0]))
                 #debug_cb("starting ssh: %s with kwargs=%s" % (str(cmd), kwargs))
@@ -1213,10 +1366,10 @@ def connect_to(display_desc, debug_cb=None, ssh_fail_cb=ssh_connect_failed):
             #instead of lazily converting everything to a string, we validate the command:
             for x in cmd:
                 if type(x)!=str:
-                    raise Exception("argument is not a string: %s (%s), found in command: %s" % (x, type(x), cmd))
+                    raise InitException("argument is not a string: %s (%s), found in command: %s" % (x, type(x), cmd))
             child = Popen(cmd, stdin=PIPE, stdout=PIPE, **kwargs)
         except OSError as e:
-            raise Exception("Error running ssh program '%s': %s" % (cmd, e))
+            raise InitException("Error running ssh program '%s': %s" % (cmd, e))
         def abort_test(action):
             """ if ssh dies, we don't need to try to read/write from its sockets """
             e = child.poll()
@@ -1231,11 +1384,11 @@ def connect_to(display_desc, debug_cb=None, ssh_fail_cb=ssh_connect_failed):
                     from xpra.log import Logger
                     log = Logger()
                     if conn.input_bytecount==0 and conn.output_bytecount==0:
-                        cmd_info = " ".join(display_desc["full_ssh"])
                         log.error("Connection to the xpra server via SSH failed for: %s", display_name)
                         log.error(" check your username, hostname, display number, firewall, etc")
                     else:
                         log.error("The SSH process has terminated with exit code %s", e)
+                    cmd_info = " ".join(display_desc["full_ssh"])
                     log.error(" the command line used was:")
                     log.error(" %s", cmd_info)
                 raise ConnectionClosedException(error_message)
@@ -1282,6 +1435,33 @@ def connect_to(display_desc, debug_cb=None, ssh_fail_cb=ssh_connect_failed):
             raise InitException("cannot connect to %s: %s" % (sockpath, e))
         conn.timeout = SOCKET_TIMEOUT
 
+    elif dtype == "named-pipe":
+        pipe_name = display_desc["named-pipe"]
+        if not sys.platform.startswith("win"):
+            raise InitException("named pipes are only supported on MS Windows")
+        import errno
+        from xpra.platform.win32.dotxpra import PIPE_PATH
+        from xpra.platform.win32.namedpipes.connection import NamedPipeConnection
+        from win32file import CreateFile, GENERIC_READ, GENERIC_WRITE, OPEN_EXISTING    #@UnresolvedImport
+        path = PIPE_PATH+pipe_name
+        try:
+            pipe_handle = CreateFile(path, GENERIC_READ | GENERIC_WRITE, 0, None, OPEN_EXISTING, 0, None)
+        except Exception as e:
+            if e[0]==errno.ENOENT:
+                raise InitException("the named pipe '%s' does not exist" % pipe_name)
+            raise InitException("failed to connect to the named pipe '%s': %s" % (pipe_name, e))
+        conn = NamedPipeConnection(pipe_name, pipe_handle)
+        conn.timeout = SOCKET_TIMEOUT
+
+    elif dtype == "vsock":
+        cid, iport = display_desc["vsock"]
+        from xpra.net.vsock import connect_vsocket, CID_TYPES       #@UnresolvedImport
+        sock = connect_vsocket(cid=cid, port=iport)
+        sock.timeout = VSOCK_TIMEOUT
+        sock.settimeout(None)
+        from xpra.net.bytestreams import SocketConnection
+        return SocketConnection(sock, "local", "host", (CID_TYPES.get(cid, cid), iport), dtype)
+
     elif dtype == "tcp":
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         sock.settimeout(SOCKET_TIMEOUT)
@@ -1306,6 +1486,7 @@ def get_sockpath(display_desc, error_cb):
     return sockpath
 
 def run_client(error_cb, opts, extra_args, mode):
+    validate_encryption(opts)
     if mode=="screenshot":
         if len(extra_args)==0:
             error_cb("invalid number of arguments for screenshot mode")
@@ -1359,14 +1540,17 @@ def run_client(error_cb, opts, extra_args, mode):
         except RuntimeError as e:
             #exceptions at this point are still initialization exceptions
             raise InitException(e.message)
+        ehelp = "help" in opts.encodings
+        if ehelp:
+            from xpra.codecs.loader import PREFERED_ENCODING_ORDER
+            opts.encodings = PREFERED_ENCODING_ORDER
         app.init(opts)
-        if opts.encoding:
-            #fix old encoding names if needed:
-            err = opts.encoding not in app.get_encodings()
+        if opts.encoding or ehelp:
+            err = opts.encoding and (opts.encoding not in app.get_encodings())
             info = ""
             if err and opts.encoding!="help":
                 info = "invalid encoding: %s\n" % opts.encoding
-            if opts.encoding=="help" or err:
+            if opts.encoding=="help" or ehelp or err:
                 from xpra.codecs.loader import encodings_help
                 raise InitInfo(info+"%s xpra client supports the following encodings:\n * %s" % (app.client_toolkit(), "\n * ".join(encodings_help(app.get_encodings()))))
         def handshake_complete(*args):
@@ -1378,6 +1562,9 @@ def run_client(error_cb, opts, extra_args, mode):
         app.init_ui(opts, extra_args)
         try:
             conn = connect()
+            #UGLY warning: connect will parse the display string,
+            #which may change the username..
+            app.username = opts.username
             app.setup_connection(conn)
         except Exception as e:
             app.cleanup()
@@ -1512,6 +1699,22 @@ def no_gtk():
     raise Exception("the gtk module is already loaded: %s" % gtk)
 
 
+def run_glcheck(opts):
+    from xpra.util import pver
+    from xpra.client.gl.gl_check import check_support
+    try:
+        props = check_support(force_enable=opts.opengl)
+    except Exception as e:
+        sys.stdout.write("error=%s\n" % e)
+        return 1
+    for k in sorted(props.keys()):
+        v = props[k]
+        #skip not human readable:
+        if k not in ("extensions", "glconfig"):
+            sys.stdout.write("%s=%s\n" % (str(k), pver(v)))
+    return 0
+
+
 def run_proxy(error_cb, opts, script_file, args, mode, defaults):
     from xpra.scripts.fdproxy import XpraProxy
     no_gtk()
@@ -1536,9 +1739,13 @@ def run_proxy(error_cb, opts, script_file, args, mode, defaults):
             assert len(args) in (0, 1), "_shadow_start: expected 0 or 1 arguments but got %s: %s" % (len(args), args)
             cmd.append("shadow")
             display_name = None
-            if len(args)==1 and args[0]==":":
+            OSXWIN = sys.platform.startswith("darwin") or sys.platform.startswith("win")
+            if len(args)==1 and (args[0]==":" or OSXWIN):
                 #display_name was provided:
                 display_name = args[0]
+            elif OSXWIN:
+                #no need for a specific display
+                display_name = "MAIN"
             else:
                 display_name = guess_X11_display(opts.socket_dir, opts.socket_dirs)
                 #we now know the display name, so add it:
@@ -1708,7 +1915,7 @@ def run_list(error_cb, opts, extra_args):
                 unknown.append((socket_dir, display, sockpath))
     #now, re-probe the "unknown" ones:
     if len(unknown)>0:
-        sys.stdout.write("Re-probing unknown sessions: %s\n" % [x[0] for x in unknown])
+        sys.stdout.write("Re-probing unknown sessions in: %s\n" % csv(list(set([x[0] for x in unknown]))))
     counter = 0
     while len(unknown)>0 and counter<5:
         time.sleep(1)
