@@ -7,9 +7,14 @@
 import sys
 import os
 
-from xpra.util import csv
+from xpra.util import csv, engs, parse_simple_dict
 from xpra.log import Logger
-log = Logger("sound")
+log = Logger("sound", "gstreamer")
+
+
+#used on the server (reversed):
+XPRA_PULSE_SOURCE_DEVICE_NAME = "Speaker"
+XPRA_PULSE_SINK_DEVICE_NAME = "Microphone"
 
 GST_QUEUE_NO_LEAK             = 0
 GST_QUEUE_LEAK_UPSTREAM       = 1
@@ -32,9 +37,8 @@ WIN32 = sys.platform.startswith("win")
 OSX = sys.platform.startswith("darwin")
 
 ALLOW_SOUND_LOOP = os.environ.get("XPRA_ALLOW_SOUND_LOOP", "0")=="1"
-DEFAULT_GSTREAMER1 = not WIN32
-GSTREAMER1 = os.environ.get("XPRA_GSTREAMER1", str(int(DEFAULT_GSTREAMER1)))=="1"
-INPUT_DEVICE_NAME = os.environ.get("XPRA_INPUT_DEVICE_NAME", "")
+GSTREAMER1 = os.environ.get("XPRA_GSTREAMER1", "1")=="1"
+PULSEAUDIO_DEVICE_NAME = os.environ.get("XPRA_PULSEAUDIO_DEVICE_NAME", "")
 def force_enabled(codec_name):
     return os.environ.get("XPRA_SOUND_CODEC_ENABLE_%s" % codec_name.upper(), "0")=="1"
 
@@ -50,6 +54,7 @@ NAME_TO_SRC_PLUGIN = {
     "pulse"         : "pulsesrc",
     "direct"        : "directsoundsrc",
     }
+SRC_HAS_DEVICE_NAME = ["alsasrc", "osssrc", "oss4src", "jackaudiosrc", "pulsesrc", "directsoundsrc", "osxaudiosrc"]
 SRC_TO_NAME_PLUGIN = {}
 for k,v in NAME_TO_SRC_PLUGIN.items():
     SRC_TO_NAME_PLUGIN[v] = k
@@ -80,28 +85,52 @@ OPUS = "opus"
 SPEEX = "speex"
 WAVPACK = "wavpack"
 
+GDP = "gdp"
+OGG = "ogg"
+MKA = "mka"
+#RTP = "rtp"
+
+FLAC_GDP    = FLAC+"+"+GDP
+OPUS_GDP    = OPUS+"+"+GDP
+SPEEX_GDP   = SPEEX+"+"+GDP
+VORBIS_OGG  = VORBIS+"+"+OGG
+#OPUS_WEBM   = OPUS+"+"+WEBM
+#OPUS_RTP    = OPUS+"+"+RTP
+VORBIS_MKA = VORBIS+"+"+MKA
+
 #format: encoder, container-formatter, decoder, container-parser
 #we keep multiple options here for the same encoding
 #and will populate the ones that are actually available into the "CODECS" dict
 CODEC_OPTIONS = [
-            (VORBIS      , "vorbisenc",     "gdppay",   "vorbisdec",    "gdpdepay"),
-            (FLAC        , "flacenc",       "oggmux",   "flacdec",      "oggdemux"),
-#            (FLAC        , "flacenc",       "gdppay",   "flacdec",      "gdpdepay"),
-            (MP3         , "lamemp3enc",    None,       "mad",          "mp3parse"),
-            (MP3         , "lamemp3enc",    None,       "mad",          "mpegaudioparse"),
-            (WAV         , "wavenc",        None,       None,           "wavparse"),
-            (OPUS        , "opusenc",       "oggmux",   "opusdec",      "oggdemux"),
-#            (OPUS        , "opusenc",       "gdppay",   "opusdec",      "gdpdepay"),
-            (SPEEX       , "speexenc",      "oggmux",   "speexdec",     "oggdemux"),
-#            (SPEEX       , "speexenc",      "gdppay",   "speexdec",     "gdpdepay"),
-            (WAVPACK     , "wavpackenc",    None,       "wavpackdec",   "wavpackparse"),
+        (VORBIS     , "vorbisenc",      "gdppay",       "vorbisdec",    "gdpdepay"),
+        (VORBIS_MKA , "vorbisenc",      "webmmux",      "vorbisdec",    "matroskademux"),
+        #fails silently - no idea why:
+        #(VORBIS_OGG , "vorbisenc",      "oggmux",       "vorbisparse ! vorbisdec",    "oggdemux"),
+        #does not work - no idea why:
+        #(FLAC       , "flacenc",        "oggmux",       "flacparse ! flacdec",      "oggdemux"),
+        #this only works in gstreamer 0.10 and is filtered out during initialization:
+        (FLAC       , "flacenc",        "oggmux",       "flacdec",      "oggdemux"),
+        (FLAC_GDP   , "flacenc",        "gdppay",       "flacparse ! flacdec",      "gdpdepay"),
+        (MP3        , "lamemp3enc",     None,           "mad",          "mp3parse"),
+        (MP3        , "lamemp3enc",     None,           "mad",          "mpegaudioparse"),
+        (WAV        , "wavenc",         None,           None,           "wavparse"),
+        (OPUS       , "opusenc",        "oggmux",       "opusdec",      "oggdemux"),
+        (OPUS_GDP   , "opusenc",        "gdppay",       "opusdec",      "gdpdepay"),
+        #for rtp, we would need to send the caps:
+        #(OPUS_RTP   , "opusenc",        "rtpopuspay",   "opusdec",      "rtpopusdepay"),
+        #(OPUS_RTP   , "opusenc",        "rtpopuspay",   "opusparse ! opusdec",      "rtpopusdepay"),
+        #this causes "could not link opusenc0 to webmmux0"
+        #(OPUS_WEBM  , "opusenc",        "webmmux",      "opusdec",      "matroskademux"),
+        #(OPUS_WEBM  , "opusenc",        "webmmux",      "opusparse ! opusdec",      "matroskademux"),
+        (SPEEX      , "speexenc",       "oggmux",       "speexdec",     "oggdemux"),
+        (SPEEX_GDP  , "speexenc",       "gdppay",       "speexdec",     "gdpdepay"),
+        (WAVPACK    , "wavpackenc",      None,          "wavpackdec",   "wavpackparse"),
             ]
 
-GDP = "gdp"
-OGG = "ogg"
 MUX_OPTIONS = [
                (GDP,    "gdppay",   "gdpdepay"),
                (OGG,    "oggmux",   "oggdemux"),
+               (MKA,    "webmmux",  "matroskademux"),
               ]
 emux = [x for x in os.environ.get("XPRA_MUXER_OPTIONS", "").split(",") if len(x.strip())>0]
 if emux:
@@ -121,12 +150,26 @@ SOURCE_NEEDS_AUDIOCONVERT = ("directsoundsrc", "osxaudiosrc", "autoaudiosrc")
 
 #options we use to tune for low latency:
 OGG_DELAY = 20*MS_TO_NS
-ENCODER_DEFAULT_OPTIONS = {
+ENCODER_DEFAULT_OPTIONS_COMMON = {
             "lamemp3enc"    : {"encoding-engine-quality": 0},   #"fast"
             "wavpackenc"    : {"mode" : 1},     #"fast" (0 aka "very fast" is not supported)
             "flacenc"       : {"quality" : 0},  #"fast"
-            "opusenc"       : {"cbr" : 0,
-                               "complexity" : 0},
+                           }
+ENCODER_DEFAULT_OPTIONS = {
+                            0       : {
+                                       "opusenc"       : {
+                                                          "cbr"            : 0,
+                                                          "complexity"     : 0
+                                                          },
+                                       },
+                            1      :   {
+                                        #FIXME: figure out when it is safe to apply the "bitrate-type" setting:
+                                        "opusenc"       : {
+                                                           #only available with 1.6 onwards?
+                                                           #"bitrate-type"   : 2,      #constrained vbr
+                                                           "complexity"     : 0
+                                                           },
+                                        },
                            }
 #we may want to review this if/when we implement UDP transport:
 GDPPAY_CRC = False
@@ -137,19 +180,26 @@ MUXER_DEFAULT_OPTIONS = {
             "gdppay"        : {"crc-header"    : int(GDPPAY_CRC),
                                "crc-payload"   : int(GDPPAY_CRC),
                                },
+            "webmmux"       : {"writing-app"    : "Xpra"},
            }
 
 #based on the encoder options above:
 ENCODER_LATENCY = {
+        VORBIS      : 0,
+        VORBIS_OGG  : 0,
+        VORBIS_MKA  : 0,
         MP3         : 250,
         FLAC        : 50,
+        FLAC_GDP    : 50,
         WAV         : 0,
         WAVPACK     : 600,
         OPUS        : 0,
+        OPUS_GDP    : 0,
         SPEEX       : 0,
+        SPEEX_GDP   : 0,
        }
 
-CODEC_ORDER = [VORBIS, OPUS, FLAC, MP3, WAV, WAVPACK, SPEEX]    #AAC is untested
+CODEC_ORDER = [OPUS_GDP, OPUS, VORBIS, VORBIS_MKA, FLAC_GDP, FLAC, MP3, WAV, WAVPACK, SPEEX_GDP, SPEEX]
 
 
 gst = None
@@ -236,16 +286,19 @@ def import_gst():
     if has_gst is not None:
         return gst
 
-    from xpra.gtk_common.gobject_compat import is_gtk3
-    if is_gtk3():
+    PYTHON3 = sys.version_info[0]>=3
+    if PYTHON3:
         imports = [ (import_gst1,       1) ]
+    elif GSTREAMER1:
+        imports = [
+                    (import_gst1,       1),
+                    (import_gst0_10,    0),
+                  ]
     else:
         imports = [
                     (import_gst0_10,    0),
                     (import_gst1,       1),
                   ]
-        if GSTREAMER1:
-            imports.reverse()  #try gst1 first:
     errs = {}
     saved_sys_path = sys.path[:]
     saved_os_environ = os.environ.copy()
@@ -310,7 +363,7 @@ def import_gst():
     else:
         log.warn("Warning: failed to import GStreamer:")
         for vstr,e in errs.items():
-            log.warn(" %s failed with: %s", vstr, e)
+            log.warn(" GStreamer %s: %s", vstr, e)
     has_gst = gst is not None
     return gst
 
@@ -343,11 +396,25 @@ def get_all_plugin_names():
 
 def has_plugins(*names):
     allp = get_all_plugin_names()
-    missing = [name for name in names if (name is not None and name not in allp)]
+    #support names that contain a gstreamer chain, ie: "flacparse ! flacdec"
+    snames = []
+    for x in names:
+        if not x:
+            continue
+        snames += [v.strip() for v in x.split("!")]
+    missing = [name for name in snames if (name is not None and name not in allp)]
     if len(missing)>0:
         log("missing %s from %s", missing, names)
     return len(missing)==0
 
+def get_encoder_default_options(encoder):
+    global gst_major_version, ENCODER_DEFAULT_OPTIONS_COMMON, ENCODER_DEFAULT_OPTIONS
+    #strip the muxer:
+    enc = encoder.split("+")[0]
+    options = ENCODER_DEFAULT_OPTIONS_COMMON.get(enc, {}).copy()
+    options.update(ENCODER_DEFAULT_OPTIONS.get(gst_major_version).get(enc, {}))
+    return options
+
 
 CODECS = None
 def get_codecs():
@@ -363,16 +430,28 @@ def get_codecs():
             continue
         if force_enabled(encoding):
             log.info("sound codec %s force enabled", encoding)
+        elif len([x for x in elements if (x and (x.find("matroska")>=0 or x.find("gdp")>=0))])>0 and get_gst_version()<(0, 10, 36):
+            #outdated versions of gstreamer cause problems with the gdp and matroskademux muxers,
+            #the receiver may not be able to process the data
+            #and we have no way of knowing what version they have at this point, so just disable those:
+            log("avoiding %s with gdp muxer - gstreamer version %s is too old", encoding, get_gst_version())
+            continue
+        elif encoding==OPUS_GDP and get_gst_version()>=(1, 8):
+            log("avoiding %s with gstreamer version %s", encoding, get_gst_version())
+            continue
+        elif encoding in (OPUS_GDP, OPUS) and OSX:
+            log("avoiding %s on Mac OS X", encoding)
+            continue
         elif encoding==FLAC:
             #flac problems:
-            if WIN32 and gst_major_version==0 and encoding==FLAC:
+            if WIN32 and gst_major_version==0:
                 #the gstreamer 0.10 builds on win32 use the outdated oss build,
                 #which includes outdated flac libraries with known CVEs,
                 #so avoid using those:
                 log("avoiding outdated flac module (likely buggy on win32 with gstreamer 0.10)")
                 continue
             elif gst_major_version==1:
-                log("skipping flac with GStreamer 1.x to avoid obscure 'not-neogtiated' errors I do not have time for")
+                log("skipping flac with GStreamer 1.x to avoid obscure 'not-negotiated' errors I do not have time for")
                 continue
         elif encoding==OPUS:
             if gst_major_version<1:
@@ -456,17 +535,20 @@ def plugin_str(plugin, options):
 
 def get_source_plugins():
     sources = []
-    from xpra.sound.pulseaudio_util import has_pa
-    #we have to put pulsesrc first if pulseaudio is installed
-    #because using autoaudiosource does not work properly for us:
-    #it may still choose pulse, but without choosing the right device.
-    if has_pa():
-        sources.append("pulsesrc")
-    sources.append("autoaudiosrc")
+    try:
+        from xpra.sound.pulseaudio.pulseaudio_util import has_pa
+        #we have to put pulsesrc first if pulseaudio is installed
+        #because using autoaudiosource does not work properly for us:
+        #it may still choose pulse, but without choosing the right device.
+        if has_pa():
+            sources.append("pulsesrc")
+    except ImportError as e:
+        log("get_source_plugins() no pulsesrc: %s", e)
     if OSX:
         sources.append("osxaudiosrc")
     elif WIN32:
         sources.append("directsoundsrc")
+    sources.append("autoaudiosrc")
     if os.name=="posix":
         sources += ["alsasrc", "jackaudiosrc",
                     "osssrc", "oss4src",
@@ -474,66 +556,131 @@ def get_source_plugins():
     sources.append("audiotestsrc")
     return sources
 
+def get_sink_plugins():
+    SINKS = []
+    if sys.platform.startswith("darwin"):
+        SINKS.append("osxaudiosink")
+    elif sys.platform.startswith("win"):
+        SINKS.append("directsoundsink")
+    SINKS.append("autoaudiosink")
+    try:
+        from xpra.sound.pulseaudio.pulseaudio_util import has_pa
+        if has_pa():
+            SINKS.append("pulsesink")
+    except ImportError as e:
+        log("get_sink_plugins() no pulsesink: %s", e)
+    if os.name=="posix":
+        SINKS += ["alsasink", "osssink", "oss4sink", "jackaudiosink"]
+    return SINKS
+
+def get_default_sink():
+    DEFAULT_SINK = os.environ.get("XPRA_SOUND_SINK")
+    if DEFAULT_SINK:
+        SINKS = get_sink_plugins()
+        if DEFAULT_SINK not in SINKS:
+            log.error("invalid default sound sink: '%s' is not in %s", DEFAULT_SINK, csv(SINKS))
+        else:
+            return DEFAULT_SINK
+    try:
+        from xpra.sound.pulseaudio.pulseaudio_util import has_pa
+        if has_pa():
+            return "pulsesink"
+    except ImportError as e:
+        log("get_default_sink() no pulsesink: %s", e)
+    SINKS = get_sink_plugins()
+    return SINKS[0]
+
 
 def get_test_defaults(*args):
     return  {"wave" : 2, "freq" : 110, "volume" : 0.4}
 
 WARNED_MULTIPLE_DEVICES = False
-def get_pulse_defaults(want_monitor_device, remote):
+def get_pulse_defaults(device_name_match=None, want_monitor_device=True, input_or_output=None, remote=None, xpra_device_name=None):
     """
         choose the device to use
     """
-    from xpra.sound.pulseaudio_util import has_pa, get_pa_device_options, get_default_sink
-    from xpra.sound.pulseaudio_util import get_pulse_server, get_pulse_id, set_source_mute
-    if not has_pa():
-        log.warn("pulseaudio is not available!")
-        return    None
+    try:
+        from xpra.sound.pulseaudio.pulseaudio_util import has_pa, get_pa_device_options, get_default_sink
+        from xpra.sound.pulseaudio.pulseaudio_util import get_pulse_server, get_pulse_id, set_source_mute, set_sink_mute
+        if not has_pa():
+            log.warn("Warning: pulseaudio is not available!")
+            return None
+    except ImportError as e:
+        log.warn("Warning: pulseaudio is not available!")
+        log.warn(" %s", e)
+        return None
     pa_server = get_pulse_server()
-    log("start sound, remote pulseaudio server=%s, local pulseaudio server=%s", remote.pulseaudio_server, pa_server)
-    #only worth comparing if we have a real server string
-    #one that starts with {UUID}unix:/..
-    if pa_server and pa_server.startswith("{") and \
-        remote.pulseaudio_server and remote.pulseaudio_server==pa_server:
-        log.error("Error: sound is disabled to prevent a sound loop")
-        log.error(" identical Pulseaudio server '%s'", pa_server)
-        return None
-    pa_id = get_pulse_id()
-    log("start sound, client id=%s, server id=%s", remote.pulseaudio_id, pa_id)
-    if remote.pulseaudio_id and remote.pulseaudio_id==pa_id:
-        log.error("Error: sound is disabled to prevent a sound loop")
-        log.error(" identical Pulseaudio ID '%s'", pa_id)
-        return None
-    device_type_str = ["input", "monitor"][want_monitor_device]
-    devices = get_pa_device_options(want_monitor_device, not want_monitor_device)
-    log("found pulseaudio %s devices: %s", device_type_str, devices)
+    if remote:
+        log("start sound, remote pulseaudio server=%s, local pulseaudio server=%s", remote.pulseaudio_server, pa_server)
+        #only worth comparing if we have a real server string
+        #one that starts with {UUID}unix:/..
+        if pa_server and pa_server.startswith("{") and \
+            remote.pulseaudio_server and remote.pulseaudio_server==pa_server:
+            log.error("Error: sound is disabled to prevent a sound loop")
+            log.error(" identical Pulseaudio server '%s'", pa_server)
+            return None
+        pa_id = get_pulse_id()
+        log("start sound, client id=%s, server id=%s", remote.pulseaudio_id, pa_id)
+        if remote.pulseaudio_id and remote.pulseaudio_id==pa_id:
+            log.error("Error: sound is disabled to prevent a sound loop")
+            log.error(" identical Pulseaudio ID '%s'", pa_id)
+            return None
+
+    device_type_str = ""
+    if input_or_output is not None:
+        device_type_str = ["output", "input"][input_or_output]
+    if want_monitor_device:
+        device_type_str += " monitor"
+    #def get_pa_device_options(monitors=False, input_or_output=None, ignored_devices=["bell-window-system"], log_errors=True)
+    devices = get_pa_device_options(want_monitor_device, input_or_output)
+    log("found %i pulseaudio %s device%s: %s", len(devices), device_type_str, engs(devices), devices)
     if len(devices)==0:
         log.error("Error: sound forwarding is disabled")
         log.error(" could not detect any Pulseaudio %s devices", device_type_str)
         return None
-    if len(devices)>1 and INPUT_DEVICE_NAME:
-        devices = dict((k,v) for k,v in devices.items() if k.find(INPUT_DEVICE_NAME)>=0 or v.find(INPUT_DEVICE_NAME)>0)
-        if len(devices)==0:
-            log.warn("Warning: Pulseaudio %s device name filter '%s'", device_type_str, INPUT_DEVICE_NAME)
-            log.warn(" did not match any devices")
-            return None
-        elif len(devices)>1:
-            log.warn("Warning: Pulseaudio %s device name filter '%s'", device_type_str, INPUT_DEVICE_NAME)
-            log.warn(" matched %i devices", len(devices))
+    if len(devices)>1:
+        filters = []
+        matches = []
+        for match in (device_name_match, PULSEAUDIO_DEVICE_NAME, xpra_device_name):
+            if not match:
+                continue
+            if match!=xpra_device_name:
+                filters.append(match)
+            match = match.lower()
+            matches = dict((k,v) for k,v in devices.items() if k.lower().find(match)>=0 or v.lower().find(match)>=0)
+            #log("matches(%s, %s)=%s", devices, match, matches)
+            if len(matches)==1:
+                log("found name match for '%s': %s", match, matches.items()[0])
+                break
+            elif len(matches)>1:
+                log.warn("Warning: Pulseaudio %s device name filter '%s'", device_type_str, match)
+                log.warn(" matched %i devices", len(matches))
+        if filters or len(matches)>0:
+            if len(matches)==0:
+                log.warn("Warning: Pulseaudio %s device name filter%s:", device_type_str, engs(filters))
+                log.warn(" %s", csv("'%s'" % x for x in filters))
+                log.warn(" did not match the devices found:")
+                for k,v in devices.items():
+                    log.warn(" * '%s'", k)
+                    log.warn("   '%s'", v)
+                return None
+            devices = matches
     #default to first one:
     device, device_name = devices.items()[0]
-    if len(devices)>1 and want_monitor_device:
+    if len(devices)>1:
         default_sink = get_default_sink()
         default_monitor = default_sink+".monitor"
         global WARNED_MULTIPLE_DEVICES
         if not WARNED_MULTIPLE_DEVICES:
             WARNED_MULTIPLE_DEVICES = True
-            if not INPUT_DEVICE_NAME: #warned already
-                log.warn("Warning: found more than one audio monitor device:")
+            if not PULSEAUDIO_DEVICE_NAME: #warned already
+                log.warn("Warning: found %i audio devices:", len(devices))
             for k,v in devices.items():
                 log.warn(" * %s", v)
                 log.warn("   %s", k)
-            if not INPUT_DEVICE_NAME: #used already!
-                log.warn(" use the environment variable XPRA_INPUT_DEVICE_NAME to select a specific one")
+            if not PULSEAUDIO_DEVICE_NAME: #used already!
+                log.warn(" to select a specific one,")
+                log.warn(" use the environment variable XPRA_PULSEAUDIO_DEVICE_NAME")
         if default_monitor in devices:
             device = default_monitor
             device_name = devices.get(default_monitor)
@@ -545,37 +692,36 @@ def get_pulse_defaults(want_monitor_device, remote):
     log.info("using pulseaudio device:")
     log.info(" '%s'", device_name)
     #make sure it is not muted:
-    set_source_mute(device, mute=False)
+    if input_or_output is True or want_monitor_device:
+        set_source_mute(device, mute=False)
+    elif input_or_output is False:
+        set_sink_mute(device, mute=False)
     return {"device" : device}
 
+def get_pulse_source_defaults(device_name_match=None, want_monitor_device=True, remote=None):
+    return get_pulse_defaults(device_name_match, want_monitor_device, input_or_output=not want_monitor_device, remote=remote, xpra_device_name=XPRA_PULSE_SOURCE_DEVICE_NAME)
+
+def get_pulse_sink_defaults():
+    return get_pulse_defaults(want_monitor_device=False, input_or_output=False, xpra_device_name=XPRA_PULSE_SINK_DEVICE_NAME)
+
 #a list of functions to call to get the plugin options
 #at runtime (so we can perform runtime checks on remote data,
 # to avoid sound loops for example)
 DEFAULT_SRC_PLUGIN_OPTIONS = {
     "test"                  : get_test_defaults,
-    "pulse"                 : get_pulse_defaults,
+    "pulse"                 : get_pulse_source_defaults,
     }
 
+DEFAULT_SINK_PLUGIN_OPTIONS = {
+    "pulse"                 : get_pulse_sink_defaults,
+    }
 
 
-def parse_element_options(options_str):
-    #parse the options string and add the pairs:
-    options = {}
-    for s in options_str.split(","):
-        if not s:
-            continue
-        try:
-            k,v = s.split("=", 1)
-            options[k] = v
-        except Exception as e:
-            log.warn("failed to parse plugin option '%s': %s", s, e)
-    return options
-
 def format_element_options(options):
     return csv("%s=%s" % (k,v) for k,v in options.items())
 
 
-def get_sound_source_options(plugin, options_str, want_monitor_device, remote):
+def get_sound_source_options(plugin, options_str, device, want_monitor_device, remote):
     """
         Given a plugin (short name), options string and remote info,
         return the options for the plugin given,
@@ -586,17 +732,23 @@ def get_sound_source_options(plugin, options_str, want_monitor_device, remote):
     #use the defaults as starting point:
     defaults_fn = DEFAULT_SRC_PLUGIN_OPTIONS.get(plugin)
     if defaults_fn:
-        options = defaults_fn(want_monitor_device, remote)
+        options = defaults_fn(device, want_monitor_device, remote)
         if options is None:
             #means failure
             return None
     else:
         options = {}
-    options.update(parse_element_options(options_str))
+        #if we add support for choosing devices in the GUI,
+        #this code will then get used:
+        if device and plugin in SRC_HAS_DEVICE_NAME:
+            #assume the user knows the "device-name"...
+            #(since I have no idea where to get the "device" string)
+            options["device-name"] = device
+    options.update(parse_simple_dict(options_str))
     return options
 
 
-def parse_sound_source(all_plugins, sound_source_plugin, want_monitor_device, remote):
+def parse_sound_source(all_plugins, sound_source_plugin, device, want_monitor_device, remote):
     #format: PLUGINNAME:options
     #ie: test:wave=2,freq=110,volume=0.4
     #ie: pulse:device=device.alsa_input.pci-0000_00_14.2.analog-stereo
@@ -619,7 +771,7 @@ def parse_sound_source(all_plugins, sound_source_plugin, want_monitor_device, re
         log.error("unknown source plugin: '%s' / '%s'", simple_str, sound_source_plugin)
         return  None, {}
     log("parse_sound_source(%s, %s, %s) plugin=%s", all_plugins, sound_source_plugin, remote, gst_sound_source_plugin)
-    options = get_sound_source_options(simple_str, options_str, want_monitor_device, remote)
+    options = get_sound_source_options(simple_str, options_str, device, want_monitor_device, remote)
     log("get_sound_source_options%s=%s", (simple_str, options_str, remote), options)
     if options is None:
         #means error
@@ -628,7 +780,6 @@ def parse_sound_source(all_plugins, sound_source_plugin, want_monitor_device, re
 
 
 def sound_option_or_all(name, options, all_values):
-    from xpra.util import engs
     if not options:
         v = all_values        #not specified on command line: use default
     else:
@@ -663,26 +814,27 @@ def loop_warning(mode="speaker", machine_id=""):
 
 def main():
     global pygst_version, gst_version, gst_vinfo
-    from xpra.platform import init, clean
+    from xpra.platform import program_context
     from xpra.log import enable_color
-    try:
-        init("GStreamer-Info", "GStreamer Information")
+    with program_context("GStreamer-Info", "GStreamer Information"):
         enable_color()
         if "-v" in sys.argv or "--verbose" in sys.argv:
             log.enable_debug()
         import_gst()
         print("Loaded Python GStreamer version %s for Python %s.%s" % (gst_vinfo, sys.version_info[0], sys.version_info[1]))
-        print("GStreamer plugins found: %s" % ", ".join(get_all_plugin_names()))
+        apn = get_all_plugin_names()
+        print("GStreamer plugins found: %s" % csv(apn))
         print("")
         print("GStreamer version: %s" % ".".join([str(x) for x in gst_version]))
         print("PyGStreamer version: %s" % ".".join([str(x) for x in pygst_version]))
         print("")
         encs = [x for x in CODEC_ORDER if has_encoder(x)]
         decs = [x for x in CODEC_ORDER if has_decoder(x)]
-        print("encoders supported: %s" % ", ".join(encs))
-        print("decoders supported: %s" % ", ".join(decs))
-    finally:
-        clean()
+        print("encoders supported: %s" % csv(encs))
+        print("decoders supported: %s" % csv(decs))
+        print("source plugins: %s" % csv([x for x in get_source_plugins() if x in apn]))
+        print("sink plugins: %s" % csv([x for x in get_sink_plugins() if x in apn]))
+        print("default sink: %s" % get_default_sink())
 
 
 if __name__ == "__main__":
